/**
 *
 */
package org.jboss.cache;

import static org.jboss.cache.AbstractNode.NodeFlags.DELETED;
import static org.jboss.cache.AbstractNode.NodeFlags.RESIDENT;

import java.util.EnumSet;
import java.util.Map;

/**
 * Base class for {@link UnversionedNode}.
 *
 * @author manik
 */
public abstract class AbstractNode<K, V>// implements Node<K, V>
{
   protected Map<Object, Node<K, V>> children;
   protected Fqn<?> fqn;
   /**
    * Flags placed on the node.  Replaces older 'boolean' flags.
    */
   protected final EnumSet<NodeFlags> flags = EnumSet.noneOf(NodeFlags.class);

   /**
    * These flags were originally stored as booleans on the UnversionedNode class.  They have been replaced with an enum
    * and an EnumSet, which is much more space-efficient for very little cost in lookups.
    */
   public enum NodeFlags
   {
      /**
       * All children are loaded from the cache loader if this flag is present.
       */
      CHILDREN_LOADED,
      /**
       * Data is loaded from the cache loader if this flag is present.
       */
      DATA_LOADED,
      /**
       * Node is write-locked when children are added or removed if this flag is enabled.
       */
      LOCK_FOR_CHILD_INSERT_REMOVE,
      /**
       * Node is valid if this flag is present.
       */
      VALID,
      /**
       * Node has been deleted.
       */
      DELETED,
      /**
       * NOde is resident and excluded from evictions
       */
      RESIDENT,
      /**
       * Specific to Optimistic Locking Workspace nodes - set if a node has been modified in a workspace.
       */
      MODIFIED_IN_WORKSPACE,
      /**
       * Specific to Optimistic Locking Workspace nodes - set if a node has been created in a workspace.
       */
      CREATED_IN_WORKSPACE,
      /**
       * Specific to Optimistic Locking Workspace nodes - set if a node has added or removed children in a workspace.
       */
      CHILDREN_MODIFIED_IN_WORKSPACE,
      /**
       * Specific to Optimistic Locking Workspace nodes - set if an implicit version is associated with this node.
       */
      VERSIONING_IMPLICIT,
      /**
       * Specific to Optimistic Locking Workspace nodes - set if a node has been resurrected in a workspace.
       */
      RESURRECTED_IN_WORKSPACE
   }

   public boolean isDeleted()
   {
      return flags.contains(DELETED);
   }

   public void markAsDeleted(boolean marker)
   {
      markAsDeleted(marker, false);
   }

   public void markAsDeleted(boolean marker, boolean recursive)
   {
      setFlag(DELETED, marker);

      if (recursive && children != null)
      {
         synchronized (this)
         {
            for (Node<?, ?> child : children.values())
            {
               ((NodeSPI) child).markAsDeleted(marker, true);
            }
         }
      }
   }

   public void setResident(boolean resident)
   {
      setFlag(RESIDENT, resident);
   }


   public boolean isResident()
   {
      return flags.contains(RESIDENT);
   }

   @Override
   public boolean equals(Object another)
   {
      if (another instanceof AbstractNode)
      {
         AbstractNode<?, ?> anotherNode = (AbstractNode) another;
         return fqn == null && anotherNode.fqn == null || !(fqn == null || anotherNode.fqn == null) && fqn.equals(anotherNode.fqn);
      }
      return false;
   }

   @Override
   public int hashCode()
   {
      return fqn.hashCode();
   }

   /**
    * Utility method for setting or unsetting a flag.  If status is true, the NodeFlag specified is added to the {@link #flags}
    * EnumSet.  If status is false, the NodeFlag is removed from the EnumSet.
    *
    * @param flag   flag to set or unset
    * @param status true to set the flag, false to unset the flag.
    */
   protected void setFlag(NodeFlags flag, boolean status)
   {
      if (status)
         flags.add(flag);
      else
         flags.remove(flag);
   }
}
