package org.jboss.cache.buddyreplication;

import org.jboss.cache.CacheException;
import org.jboss.cache.Fqn;
import org.jboss.cache.InvocationContext;
import org.jboss.cache.commands.AbstractVisitor;
import org.jboss.cache.commands.ReversibleCommand;
import org.jboss.cache.commands.read.ExistsCommand;
import org.jboss.cache.commands.read.GetChildrenNamesCommand;
import org.jboss.cache.commands.read.GetDataMapCommand;
import org.jboss.cache.commands.read.GetKeyValueCommand;
import org.jboss.cache.commands.read.GetKeysCommand;
import org.jboss.cache.commands.read.GetNodeCommand;
import org.jboss.cache.commands.read.GravitateDataCommand;
import org.jboss.cache.commands.tx.CommitCommand;
import org.jboss.cache.commands.tx.OptimisticPrepareCommand;
import org.jboss.cache.commands.tx.PrepareCommand;
import org.jboss.cache.commands.tx.RollbackCommand;
import org.jboss.cache.commands.write.*;
import org.jboss.cache.factories.CommandsFactory;
import org.jgroups.Address;

import java.util.ArrayList;
import java.util.List;

/**
 * For each command the fqns are chnaged such that they are under the current buddy group's backup subtree
 * (e.g., /_buddy_backup_/my_host:7890/) rather than the root (/).
 *
 * @author Mircea.Markus@jboss.com
 * @since 2.2
 */
public class BuddyFqnTransformer extends AbstractVisitor
{


   public static final String BUDDY_BACKUP_SUBTREE = "_BUDDY_BACKUP_";
   public static final Fqn<String> BUDDY_BACKUP_SUBTREE_FQN = Fqn.fromString(BUDDY_BACKUP_SUBTREE);


   private final String buddyGroupName;

   CommandsFactory factory;

   public BuddyFqnTransformer(String buddyGroupName)
   {
      this.buddyGroupName = buddyGroupName == null ? "null" : buddyGroupName;
   }

   public BuddyFqnTransformer(String buddyGroupName, CommandsFactory cf)
   {
      this.buddyGroupName = buddyGroupName == null ? "null" : buddyGroupName;
      this.factory = cf;
   }

   @Override
   public Object visitCommitCommand(InvocationContext ctx, CommitCommand commitCommand) throws Throwable
   {
      return commitCommand;
   }

   @Override
   public Object visitPutDataMapCommand(InvocationContext ctx, PutDataMapCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildPutDataMapCommand(null, transformed, command.getData());
   }

   @Override
   public Object visitPutKeyValueCommand(InvocationContext ctx, PutKeyValueCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildPutKeyValueCommand(null, transformed, command.getKey(), command.getValue());
   }

   @Override
   public Object visitPutForExternalReadCommand(InvocationContext ctx, PutForExternalReadCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildPutForExternalReadCommand(null, transformed, command.getKey(), command.getValue());
   }

   @Override
   public Object visitRemoveNodeCommand(InvocationContext ctx, RemoveNodeCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildRemoveNodeCommand(command.getGlobalTransaction(), transformed);
   }

   @Override
   public Object visitClearDataCommand(InvocationContext ctx, ClearDataCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildClearDataCommand(command.getGlobalTransaction(), transformed);
   }

   @Override
   public Object visitEvictFqnCommand(InvocationContext ctx, EvictCommand command) throws Throwable
   {
      Fqn fqn = getBackupFqn(command.getFqn());
      return factory.buildEvictFqnCommand(fqn);
   }

   @Override
   public Object visitInvalidateCommand(InvocationContext ctx, InvalidateCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildInvalidateCommand(transformed);
   }

   @Override
   public Object visitRemoveKeyCommand(InvocationContext ctx, RemoveKeyCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildRemoveKeyCommand(null, transformed, command.getKey());
   }

   @Override
   public Object visitGetDataMapCommand(InvocationContext ctx, GetDataMapCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildGetDataMapCommand(transformed);
   }

   @Override
   public Object visitExistsNodeCommand(InvocationContext ctx, ExistsCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildEvictFqnCommand(transformed);
   }

   @Override
   public Object visitGetKeyValueCommand(InvocationContext ctx, GetKeyValueCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildGetKeyValueCommand(transformed, command.getKey(), command.isSendNodeEvent());
   }

   @Override
   public Object visitGetNodeCommand(InvocationContext ctx, GetNodeCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildGetNodeCommand(transformed);
   }

   @Override
   public Object visitGetKeysCommand(InvocationContext ctx, GetKeysCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildGetKeysCommand(transformed);
   }

   @Override
   public Object visitGetChildrenNamesCommand(InvocationContext ctx, GetChildrenNamesCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildGetChildrenNamesCommand(transformed);
   }

   @Override
   public Object visitMoveCommand(InvocationContext ctx, MoveCommand command) throws Throwable
   {
      Fqn transformedFrom = getBackupFqn(command.getFqn());
      Fqn transformedTo = getBackupFqn(command.getTo());
      return factory.buildMoveCommand(transformedFrom, transformedTo);
   }

   @Override
   public Object visitGravitateDataCommand(InvocationContext ctx, GravitateDataCommand command) throws Throwable
   {
      Fqn transformed = getBackupFqn(command.getFqn());
      return factory.buildGravitateDataCacheCommand(transformed, command.isSearchSubtrees());
   }

   @Override
   public Object visitPrepareCommand(InvocationContext ctx, PrepareCommand command) throws Throwable
   {
      List<ReversibleCommand> toTransform = command.getModifications();
      List<ReversibleCommand> transformedCommands = transformBatch(toTransform);
      return factory.buildPrepareCommand(command.getGlobalTransaction(), transformedCommands, command.getLocalAddress(), command.isOnePhaseCommit());
   }

   @Override
   public Object visitRollbackCommand(InvocationContext ctx, RollbackCommand command) throws Throwable
   {
      return factory.buildRollbackCommand(null);
   }

   @Override
   public Object visitOptimisticPrepareCommand(InvocationContext ctx, OptimisticPrepareCommand command) throws Throwable
   {
      List<ReversibleCommand> transformed = transformBatch(command.getModifications());
      return factory.buildOptimisticPrepareCommand(command.getGlobalTransaction(), transformed, command.getData(), command.getLocalAddress(), command.isOnePhaseCommit());
   }

   @Override
   public Object visitCreateNodeCommand(InvocationContext ctx, CreateNodeCommand command) throws Throwable
   {
      return factory.buildCreateNodeCommand(getBackupFqn(command.getFqn()));
   }

   /**
    * Assumes the backup Fqn if the current instance is the data owner.
    */
   public Fqn getBackupFqn(Fqn originalFqn)
   {
      return getBackupFqn(buddyGroupName, originalFqn);
   }


   public List<ReversibleCommand> transformBatch(List<ReversibleCommand> toTransform) throws Throwable
   {
      List<ReversibleCommand> transformedCommands = new ArrayList<ReversibleCommand>(toTransform.size());
      for (ReversibleCommand com : toTransform)
      {
         transformedCommands.add((ReversibleCommand) com.acceptVisitor(null, this));
      }
      return transformedCommands;
   }

   /**
    * Utility method that retrieves a buddy backup Fqn given the actual Fqn of some data and the data owner's Address.
    *
    * @param dataOwnerAddress the JGroups {@link org.jgroups.Address}  of the data owner
    * @param origFqn          the original Fqn
    * @return a backup Fqn
    */
   public static Fqn getBackupFqn(Address dataOwnerAddress, Fqn origFqn)
   {
      return getBackupFqn(getGroupNameFromAddress(dataOwnerAddress), origFqn);
   }

   /**
    * Utility method that retrieves a buddy backup Fqn given the actual Fqn of some data and the buddy group name.
    *
    * @param buddyGroupName the buddy group name
    * @param origFqn        the original Fqn
    * @return a backup Fqn
    */
   public static Fqn getBackupFqn(String buddyGroupName, Fqn origFqn)
   {
      if (isBackupFqn(origFqn))
         throw new CacheException("Cannot make a backup Fqn from a backup Fqn! Attempting to create a backup of " + origFqn);

      List<Object> elements = new ArrayList<Object>(origFqn.size() + 2);
      elements.add(BuddyManager.BUDDY_BACKUP_SUBTREE);
      elements.add(buddyGroupName);
      elements.addAll(origFqn.peekElements());

      return Fqn.fromList(elements);
   }

   /**
    * Utility method that retrieves a buddy backup Fqn given the actual Fqn of some data and the backup subtree for the
    * buddy group in question
    *
    * @param buddyGroupRoot the subtree under which data for a particular buddy is backed up
    * @param origFqn        the original Fqn
    * @return a backup Fqn
    */
   public static Fqn getBackupFqn(Fqn buddyGroupRoot, Fqn origFqn)
   {
      if (isBackupFqn(origFqn))
         throw new CacheException("Cannot make a backup Fqn from a backup Fqn! Attempting to create a backup of " + origFqn);

      List<Object> elements = new ArrayList<Object>(origFqn.size() + 2);
      elements.add(BuddyManager.BUDDY_BACKUP_SUBTREE);
      elements.add(buddyGroupRoot.get(1));
      elements.addAll(origFqn.peekElements());

      return Fqn.fromList(elements);
   }

   public static boolean isBackupFqn(Fqn name)
   {
      return name != null && name.hasElement(BuddyManager.BUDDY_BACKUP_SUBTREE);
   }

   public static Fqn getActualFqn(Fqn fqn)
   {
      if (!isBackupFqn(fqn)) return fqn;
      // remove the first 2 (or 3 in the case of a dead backup region) elements
      return fqn.getSubFqn(isDeadBackupFqn(fqn) ? 3 : 2, fqn.size());
   }

   /**
    * Tests whether a given Fqn belongs to a dead backup region.
    *
    * @param name fqn to test
    * @return true if the fqn is a part of a dead backup region; false otherwise.
    */
   @SuppressWarnings("unchecked")
   public static boolean isDeadBackupFqn(Fqn name)
   {
      if (name == null) return false;
      Object elem1 = name.get(1);
      if (elem1 instanceof String)
      {
         String strElem1 = (String) elem1;
         return name.hasElement(BuddyManager.BUDDY_BACKUP_SUBTREE) && strElem1.endsWith(":DEAD");
      }
      else
      {
         return false;
      }
   }

   /**
    * @param dataOwner owner of a data set
    * @return a backup root for a given data owner
    */
   @SuppressWarnings("unchecked")
   public static Fqn<String> getBackupRoot(Address dataOwner)
   {
      return (Fqn) Fqn.fromRelativeElements(BUDDY_BACKUP_SUBTREE_FQN, getGroupNameFromAddress(dataOwner));
   }

   /**
    * Returns the backup root of a dead data owner
    *
    * @param dataOwner owner of data
    * @return Fqn of dead data owner's root
    */
   @SuppressWarnings("unchecked")
   public static Fqn<String> getDeadBackupRoot(Address dataOwner)
   {
      return (Fqn) Fqn.fromRelativeElements(BUDDY_BACKUP_SUBTREE_FQN, getGroupNameFromAddress(dataOwner) + ":DEAD");
   }

   public static boolean isDeadBackupRoot(Fqn f)
   {
      return f.getParent().equals(BUDDY_BACKUP_SUBTREE_FQN) && f.getLastElementAsString().endsWith(":DEAD");
   }

   public static String getGroupNameFromAddress(Address address)
   {
      return address.toString().replace(':', '_');
   }

   /**
    * Returns the buddy backp root portion of a given Fqn, provided it is a backup Fqn.  If it is not a backup Fqn, Fqn.ROOT is returned.
    *
    * @param fqn fqn
    */
   public static Fqn getBackupRootFromFqn(Fqn fqn)
   {
      if (isBackupFqn(fqn))
      {
         return fqn.getSubFqn(0, isDeadBackupFqn(fqn) ? 3 : 2);
      }
      else
      {
         return Fqn.ROOT;
      }
   }
}
