/*
 * JBoss, Home of Professional Open Source.
 * Copyright 2006, Red Hat Middleware LLC, and individual contributors
 * as indicated by the @author tags. See the copyright.txt file in the
 * distribution for a full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */

package org.jboss.cache.passivation;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.cache.CacheFactory;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.DefaultCacheFactory;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.util.TestingUtil;
import org.jboss.cache.xml.XmlHelper;
import org.jboss.cache.config.Configuration;
import org.jboss.cache.config.EvictionConfig;
import org.jboss.cache.config.CacheLoaderConfig;
import org.jboss.cache.factories.XmlConfigurationParser;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import org.jboss.cache.loader.DummyInMemoryCacheLoader;
import org.jboss.cache.notifications.annotation.CacheListener;
import org.jboss.cache.notifications.annotation.NodeActivated;
import org.jboss.cache.notifications.annotation.NodeLoaded;
import org.jboss.cache.notifications.annotation.NodePassivated;
import org.jboss.cache.notifications.event.NodeEvent;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.fail;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.w3c.dom.Element;

/**
 * @author Ben Wang, Feb 11, 2004
 */
@Test(groups = {"functional", "jgroups"})
public class ReplicatedPassivationIntegrationTest
{
   private CacheSPI<String, String> cache_;
   private CacheSPI<String, String> cache1_;
   protected final static Log log = LogFactory.getLog(ReplicatedPassivationIntegrationTest.class);
   int wakeupIntervalMillis_ = 0;
   PassivationListener listener_;

   public ReplicatedPassivationIntegrationTest()
   {
      listener_ = new ReplicatedPassivationIntegrationTest.PassivationListener();
   }

   @BeforeMethod(alwaysRun = true)
   public void setUp() throws Exception
   {
      CacheFactory<String, String> instance = new DefaultCacheFactory();
      cache_ = (CacheSPI<String, String>) instance.createCache(getCfg(), false);
      cache_.getConfiguration().setUseRegionBasedMarshalling(true);
      cache_.start();

      cache1_ = (CacheSPI<String, String>) instance.createCache(getCfg(), false);
      cache1_.getConfiguration().setUseRegionBasedMarshalling(true);

      cache1_.start();
      cache1_.getNotifier().addCacheListener(listener_);
      listener_.resetCounter();

      wakeupIntervalMillis_ = cache1_.getConfiguration().getEvictionConfig().getWakeupIntervalSeconds() * 1000;
      log("wakeupInterval is " + wakeupIntervalMillis_);
      if (wakeupIntervalMillis_ <= 0)
      {
         fail("testEviction(): eviction thread wake up interval is illegal " + wakeupIntervalMillis_);
      }
   }

   Configuration getCfg() throws Exception
   {
      Configuration cfg = UnitTestCacheConfigurationFactory.createConfiguration(Configuration.CacheMode.REPL_SYNC);
      cfg.setEvictionConfig(buildEvictionConfig());
      cfg.setCacheLoaderConfig(buildCacheLoaderConfig());
      cfg.setTransactionManagerLookupClass("org.jboss.cache.transaction.DummyTransactionManagerLookup");
      cfg.getCacheLoaderConfig().getFirstCacheLoaderConfig().setClassName(DummyInMemoryCacheLoader.class.getName());
      return cfg;
   }

   private CacheLoaderConfig buildCacheLoaderConfig() throws Exception
   {
      String xml = " <attribute name=\"CacheLoaderConfiguration\">\n" +
            "         <config>\n" +
            "            <passivation>true</passivation>\n" +
            "            <preload>/</preload>\n" +
            "            <shared>false</shared>\n" +
            "            <cacheloader>\n" +
            "               <class>org.jboss.cache.loader.FileCacheLoader</class>\n" +
            "               <properties>\n" +
            "               </properties>\n" +
            "               <async>false</async>\n" +
            "               <fetchPersistentState>true</fetchPersistentState>\n" +
            "               <ignoreModifications>false</ignoreModifications>\n" +
            "            </cacheloader>\n" +
            "         </config>\n" +
            "      </attribute>";
      Element element = XmlHelper.stringToElement(xml);
      return XmlConfigurationParser.parseCacheLoaderConfig(element);

   }

   private EvictionConfig buildEvictionConfig() throws Exception
   {
      String xml = "      <attribute name=\"EvictionPolicyConfig\">\n" +
            "         <config>\n" +
            "            <attribute name=\"wakeUpIntervalSeconds\">5</attribute>\n" +
            "            <attribute name=\"eventQueueSize\">200000</attribute>\n" +
            "            <attribute name=\"policyClass\">org.jboss.cache.eviction.LRUPolicy</attribute>\n" +
            "            <region name=\"/_default_\">\n" +
            "               <attribute name=\"maxNodes\">5000</attribute>\n" +
            "               <attribute name=\"timeToLiveSeconds\">3</attribute>\n" +
            "            </region>\n" +
            "            <region name=\"/org/jboss/test/data\">\n" +
            "               <attribute name=\"maxNodes\">100</attribute>\n" +
            "               <attribute name=\"timeToLiveSeconds\">3</attribute>\n" +
            "            </region>\n" +
            "         </config>\n" +
            "      </attribute>";
      Element element = XmlHelper.stringToElement(xml);
      return XmlConfigurationParser.parseEvictionConfig(element);

   }

   @AfterMethod(alwaysRun = true)
   public void tearDown() throws Exception
   {
      cache_.stop();
      cache1_.stop();
   }

   /**
    */
   public void testActivationEvent() throws Exception
   {
      String rootStr = "/org/jboss/test/data/";
      String rootStr1 = "/__JBossInternal__/5c4o12-pzhlhj-esnuy3sg-1-esnuy3sg-2";
      String str = rootStr + "0";
      cache_.removeNode(Fqn.ROOT);

      cache_.put(str, str, str);
      cache_.put(rootStr1, str, str);

      TestingUtil.sleepThread(11000);
      Node n = cache1_.peek(Fqn.fromString(str), false, false);
      assert n == null || !n.getKeys().contains(str) : "UnversionedNode should not exist";
      String val;
      val = cache1_.get(str, str);
      val = cache1_.get(rootStr1, str);
      assertNotNull("DataNode should be activated ", val);
   }

   void log(String msg)
   {
      System.out.println("-- " + msg);
   }

   @CacheListener
   public class PassivationListener
   {
      int counter = 0;
      int loadedCounter = 0;

      public int getCounter()
      {
         return counter;
      }

      public void resetCounter()
      {
         counter = 0;
         loadedCounter = 0;
      }

      @NodeActivated
      public void nodeActivated(NodeEvent ne)
      {
         if (!ne.isPre())
         {
            counter++;
            System.out.println("nodeActivate(): counter: " + counter);
            System.out.println("nodeActivate(): " + ne.getFqn());
         }
      }

      @NodePassivated
      public void nodePassivated(NodeEvent ne)
      {
         if (ne.isPre())
         {
            System.out.println("nodePassivate(): " + ne.getFqn());
         }
      }

      @NodeLoaded
      public void nodeLoaded(NodeEvent ne)
      {
         if (!ne.isPre())
         {
            loadedCounter++;
         }
      }

   }
}
